/*****************************************/
/*****************************************/
/**              CLAUSES                **/
/**                                     **/
/** Managing and solving a set of       **/
/** clauses                             **/
/**                                     **/
/** Copyright (c) 2008                  **/
/**     Stefan Stattelmann              **/
/**                                     **/
/** Author:                             **/
/**     Stefan Stattelmann              **/
/**                                     **/
/** Contact:                            **/
/**     s9ststat@stud.uni-saarland.de   **/
/*****************************************/
/*****************************************/

#include "clauses.h"

/***************/
/* Global data */
/***************/
static const void* sorting_info;    /* general purpose pointer used for sorting arrays */


/**************/
/* Functions  */
/**************/

/* forward declarations of static functions */
static int Abs (int l);
static VAR_STATE clauses_EvalLiteral (CLAUSES set, int l);
static void clauses_Sort (CLAUSES set, int var);
static int clauses_AllTrue (CLAUSES set);
static int clauses_SomeFalse (CLAUSES set);
static int clauses_UnitClause (CLAUSES set, int *clause, int *lit);
static int clauses_UndefinedVar (CLAUSES set);
static int clauses_Decide (CLAUSES set, int lit);
static void clauses_Backtrack (CLAUSES set, int pos);
static int clauses_DPLL (CLAUSES set);
static int* clauses_VarDependence (CLAUSES set, int v, int p);
static int clauses_VarRankingCmp (const void *a, const void *b);
static void clauses_VarRanking (CLAUSES set);


static int Abs (int l)
/**************************************************************
 INPUT:     integer l
 RETURNS:   avsolute value of l
***************************************************************/
{
    if (l < 0)
        return -l;

    return l;
}

void    clauses_Init (void)
/**************************************************************
 INPUT:     nothing
 RETURNS:   void
 SUMMARY:   must be here
***************************************************************/
{
}

void    clauses_Free (void)
/**************************************************************
 INPUT:     nothing
 RETURNS:   void
 SUMMARY:   must be here
***************************************************************/
{
}

CLAUSES clauses_Parse (const char* name)
/**************************************************************
 INPUT:     file name of a DIMACS CNF 
 RETURNS:   a set of clauses on success, NULL on failure
 SUMMARY:   parses the file and creates a clause set structure
***************************************************************/
{
    FILE *fp;
    int num_vars, num_clauses;
    LIST *clauses;
    CLAUSES set;

    set = NULL;

    fp = fopen (name, "r");

    /* check if opening file was successful */
    if (!fp)
    {
        fprintf (stderr, "Could not open file %s\n", name);
        return NULL;
    }

    /* parse file */
    if (parse_Parse (fp, &clauses, &num_vars, &num_clauses))
    {
        set = (CLAUSES) malloc (sizeof (CLAUSES_NODE));

        /* store information from parsing */
        set->num_vars = num_vars;
        set->num_clauses = num_clauses;
        set->clauses = clauses;

        /* do not allocate memory for solving yet */
        set->valuation = NULL;
        set->backtrack = NULL;
        set->dep = NULL;
        set->ranking = NULL;
        set->back_pos = 0;
        set->rank_pos = 0;
    }

    fclose (fp);

    return set;
}

void clauses_Delete (CLAUSES set)
/**************************************************************
 INPUT:     structure for a clause set
 SUMMARY:   delete all dynamically allocated memory of
            the structure
***************************************************************/
{
    int c;

    for (c = 0; c < set->num_clauses; c++)
        list_Delete (set->clauses[c], 0);

    free (set->clauses);
    free (set->valuation);
    free (set->backtrack);
    free (set->dep);
    free (set->ranking);
    free (set);
}

static VAR_STATE clauses_EvalLiteral (CLAUSES set, int l)
/**************************************************************
 INPUT:     structure for a clause set, integer encoded literal
 RETURNS:   value of the literal according to the current clause
            set valuation
 SUMMARY:   sort the literals in every clause (=list of
            literals) according to the valuation and the order
            VAR_TRUE < VAR_UNDEFINED < VAR_FALSE
***************************************************************/
{
    if (l < 0 && l >= -set->num_vars)
    {
        VAR_STATE v;
    
        v = set->valuation[-l];

        if (v == VAR_TRUE)
            return VAR_FALSE;
        else if (v == VAR_FALSE)
            return VAR_TRUE;
    }
    else if (l > 0 && l <= set->num_vars)
    {
        return set->valuation[l];
    }

    return VAR_UNDEFINED;
}

static void clauses_Sort (CLAUSES set, int var)
/**************************************************************
 INPUT:     structure for a clause set, variable id
 SUMMARY:   sort the literals in all clauses that depend on the
            given variable clause (=list of literals)
            according to the valuation and the order
            VAR_TRUE < VAR_UNDEFINED < VAR_FALSE
***************************************************************/
{
    int *c;

    c = clauses_VarDependence (set, var, 0);

    for (; *c >= 0; c++)
    {
        LIST tr, fa, un, it;

        tr = NULL;
        fa = NULL;
        un = NULL;
        
        for (it = set->clauses[*c]; it; it = list_Tail (it))
        {
            long literal;
            VAR_STATE v;

            literal = (long) list_Head (it);
            v = clauses_EvalLiteral (set, literal);

            switch (v)
            {
                case VAR_UNDEFINED:
                    un = list_Prepend ((void*) literal, un);
                    break;
                case VAR_FALSE:
                    fa = list_Prepend ((void*) literal, fa);
                    break;
                case VAR_TRUE:
                    tr = list_Prepend ((void*) literal, tr);
                    break;
            }
        }

        /* delete clause */
        list_Delete (set->clauses[*c], 0);

        /* reconstruct re-ordered clause */
        set->clauses[*c] = list_Concat (list_Concat (tr, un), fa);
    }
}

static int clauses_AllTrue (CLAUSES set)
/**************************************************************
 INPUT:     set of clauses including a valuation
 RETURNS:   nonzero if all clauses are true under the valuation
***************************************************************/
{
    int c;

    for (c = 0; c < set->num_clauses; c++)
    {
        long literal;
        VAR_STATE v;

        literal = (long) list_Head (set->clauses[c]);
        v = clauses_EvalLiteral (set, literal);

        /* if the first literal does not evaluate to true, none
           will, because literals are sorted */
        if (v != VAR_TRUE)
            return 0;
    }

    return 1;
}

static int clauses_SomeFalse (CLAUSES set)
/**************************************************************
 INPUT:     set of clauses including a valuation
 RETURNS:   nonzero if one clause is false under the valuation
***************************************************************/
{
    int c;

    for (c = 0; c < set->num_clauses; c++)
    {
        long literal;
        VAR_STATE v;

        literal = (long) list_Head (set->clauses[c]);
        v = clauses_EvalLiteral (set, literal);

        /* if the first literal does evaluate to false, all other
           will, because literals are sorted */
        if (v == VAR_FALSE)
            return 1;
    }

    return 0;
}

static int clauses_UnitClause (CLAUSES set, int *clause, int *lit)
/**************************************************************
 INPUT:     set of clauses including a valuation, pointers to
            variables storing which clause/literal are the unit
            clause/literal respectively
 RETURNS:   nonzero if a unit clause was found
***************************************************************/
{
    int c;

    for (c = 0; c < set->num_clauses; c++)
    {
        long literal;
        VAR_STATE v;
        LIST tail;

        literal = (long) list_Head (set->clauses[c]);
        v = clauses_EvalLiteral (set, literal);

        /* if the first literal is not undefined, this clause is true */
        if (v != VAR_UNDEFINED)
            continue;

        tail = list_Tail (set->clauses[c]);

        if (tail)
        {
            long literal2;

            literal2 = (long) list_Head (set->clauses[c]);
            v = clauses_EvalLiteral (set, literal2);

            /* if the next literal is also undefined, this is not a
               unit clause */
            if (v == VAR_UNDEFINED)
                continue;
        }

        /* either there is only one literal, or the next one is false */
        *clause = c;
        *lit = (int) literal;
        return 1;
    }

    return 0;
}

static int clauses_UndefinedVar (CLAUSES set)
/**************************************************************
 INPUT:     set of clauses including a valuation
 RETURNS:   id of an undefined variable
***************************************************************/
{
    int c;

    for (; set->rank_pos < set->num_vars; set->rank_pos++)
    {
        c = set->ranking[set->rank_pos];

        if (set->valuation [c] == VAR_UNDEFINED)
            return c;
    }

    return 0;
}

static int clauses_Decide (CLAUSES set, int lit)
/**************************************************************
 INPUT:     set of clauses including a valuation, a literal
 SUMMARY:   valuation is modified s.t. literal evaluates to true
            literal is pushed on backtracking stack
 RETURNS:   the position of the literal in the backtracking stack
***************************************************************/
{
    int pos;

    /* modifiy valuation */
    if (lit < 0)
    {
        lit = -lit;
        set->valuation[lit] = VAR_FALSE;
    }
    else
        set->valuation[lit] = VAR_TRUE;

    /* keep backtracking information */
    pos = set->back_pos++;
    set->backtrack[pos] = lit;

    /* update clauses that have changed */
    clauses_Sort (set, lit);

    return pos;
}

static void clauses_Backtrack (CLAUSES set, int pos)
/**************************************************************
 INPUT:     set of clauses including a valuation, a position
            in the backtracking array/stack
 SUMMARY:   valuation is modified s.t. variables in backtracking
            stack up to ( and including) pos evaluate to undefined
***************************************************************/
{
    int c;

    for (c = set->back_pos - 1; c >= pos; c--)
    {
        set->valuation[set->backtrack[c]] = VAR_UNDEFINED;

        /* update clauses that have changed */
        clauses_Sort (set, set->backtrack[c]);

        /* also move back in ranking, although we might move
           further back than necessary */
        if (set->rank_pos > 0)
            set->rank_pos--;
    }

    set->back_pos = pos;
}

static int clauses_DPLL (CLAUSES set)
/**************************************************************
 INPUT:     set of clauses including
 SUMMARY:   implements the DPLL algorithm recursively
***************************************************************/
{
    int clause, literal;

    if (clauses_AllTrue (set))
        return 1;
    else if (clauses_SomeFalse (set))
        return 0;
    else if (clauses_UnitClause (set, &clause, &literal))
    {
        clauses_Decide (set, literal);

        return clauses_DPLL (set);
    }
    else
    {
        int var, pos;

        var = clauses_UndefinedVar (set);

        /* not good if there is no undefined variable at this point */
        if (var == 0)
            return 0;

        pos = clauses_Decide (set, -var);

        if (clauses_DPLL (set))
            return 1;
        else
        {
            clauses_Backtrack (set, pos);
            clauses_Decide (set, var);
            return clauses_DPLL (set);
        }
    }

    return 0;
}

static int* clauses_VarDependence (CLAUSES set, int v, int p)
/**************************************************************
 INPUT:     clause set structure, variable id, array index
 RETURNS:   pointer into a -1 terminated array storing the
            clause ids for clauses that contain this variable
***************************************************************/
{
    if (v <= 0 && v > set->num_vars)
    {
        fprintf (stderr, "Illegal variable\n");
        exit (1);
    }

    return set->dep + (set->num_clauses + 1) * v + p;
}

static int clauses_VarRankingCmp (const void *a, const void *b)
/**************************************************************
 SUMMARY:   compare function for variables acording to their
            rank
 CAUTION:   sorting_info must be set to an array containing
            the ranking
***************************************************************/
{
    const int *x, *y;
    int *arr;

    x = (const int*) a;
    y = (const int*) b;

    /* global pointer must store variable ranking */
    arr = (int*) sorting_info;

    /* sort biggest rank first */
    if (arr[*x] > arr[*y])
        return -1;
    if (arr[*x] == arr[*y])
        return 0;

    return 1;
}

static void clauses_VarRanking (CLAUSES set)
/**************************************************************
 INPUT:     clause set structure 
 SUMMARY:   initializes the ranking for selection of undefined
            variables
***************************************************************/
{
    int c;
    LIST it;
    int *rank;

    /* init storage for counting how often a variable occurs*/
    rank = (int*) malloc (sizeof(int) * (set->num_vars+1));
    memset (rank, 0, sizeof(int) * (set->num_vars+1));

    /* init storage to store variable/clause dependency */
    set->dep = (int*) malloc (sizeof(int) * (set->num_vars+1) * (set->num_clauses+1));
    memset (set->dep, -1, sizeof(int) * (set->num_vars+1) * (set->num_clauses+1));

    /* init storage for variable ranking*/
    set->ranking = (int*) malloc (sizeof(int) * set->num_vars);
    for (c = 0; c < set->num_vars; c++)
        set->ranking[c] = c+1;

    /* count how often each variable occurs */
    for (c = 0; c < set->num_clauses; c++)
    {
        it = set->clauses[c];

        for (; it; it = list_Tail (it))
        {
            int v;
            long l;
            int *dep_pointer;

            l = (long) list_Head (it);
            v = Abs ((int) l);

            /* store dependency */
            dep_pointer = clauses_VarDependence (set, v, rank[v]);
            *dep_pointer = c;

            /* increase rank */
            rank[v]++;
        }
    }

    /* create ranking according to variable occurrence */
    sorting_info = rank;
    qsort (set->ranking, set->num_vars, sizeof(int), clauses_VarRankingCmp);
    set->rank_pos = 0;

    free (rank);
}

int clauses_Solve (CLAUSES set)
/**************************************************************
 INPUT:     clause set structure 
 RETURNS:   truth value of formulas according to DPLL algorithm
 SUMMARY:   
***************************************************************/
{
    /* Solve() should only be called once, hence those pointers
       are NULL, but you never know... */
    free (set->valuation);
    free (set->backtrack);
    free (set->dep);
    free (set->ranking);

    /* init valuation and backtracking stuff */
    set->valuation = (VAR_STATE*) malloc (sizeof (VAR_STATE) * (set->num_vars+1));
    memset (set->valuation, VAR_UNDEFINED, sizeof (VAR_STATE) * (set->num_vars+1));

    /* backtrack stack must be one item bigger than #variables because
       we could decide #variables+1 times before noticing the contradiction */
    set->backtrack = (int*) malloc (sizeof (int) * (set->num_vars+2));
    set->back_pos = 0;

    /* calculate dependence information */
    clauses_VarRanking (set);

    return clauses_DPLL (set);
}

void clauses_PrintValuation (CLAUSES set, FILE *fp)
/**************************************************************
 INPUT:     clause set structure, a file handle
 SUMMARY:   print the defined values of the current valuation
            stored in the structure to the file
***************************************************************/
{
    int c, val;

    if (!set->backtrack)
        return;

    for (c = 0; c <= set->num_vars; c++)
    {
        val =(int) set->valuation[c];

        switch (val)
        {
            case VAR_UNDEFINED:
                /* undefined variables are not printed */
                continue;
            case VAR_FALSE:
                val = -c;
                break;
            case VAR_TRUE:
                val = c;
                break;
            default:
                /* the valuation stores a value not defined for VAR_STATE */
                fprintf (stderr, "Internal valuation error");
                return;
        }

        fprintf (fp, " %d", val);
    }
}
