/*****************************************/
/*****************************************/
/**              PARSE                  **/
/**                                     **/
/** Parsing of DIMACS SAT cnf format    **/
/**                                     **/
/** Copyright (c) 2008                  **/
/**     Stefan Stattelmann              **/
/**                                     **/
/** Author:                             **/
/**     Stefan Stattelmann              **/
/**                                     **/
/** Contact:                            **/
/**     s9ststat@stud.uni-saarland.de   **/
/*****************************************/
/*****************************************/

#include "parse.h"

/**************/
/* Functions  */
/**************/

static void parse_Skip (FILE *fp)
/**************************************************************
 INPUT:     file handle to parsed file
 RETURNS:   void
 SUMMARY:   moves through file until a newline is encounterd
***************************************************************/
{
    int c, d;

    c = 0;

    do
    {
        d = c;
        c = fgetc (fp);
    }
    while (c != '\n');
}

static int parse_Header (FILE *fp, int *vars, int *clauses)
/**************************************************************
 INPUT:     file handle to parsed file and pointer to the
            variables where the header information should be
            stored
 RETURNS:   zero if operation was not successful
 SUMMARY:   get information from file header
***************************************************************/
{
    char format[128];       /* buffer for file format string, must be "cnf" */
    int res;

    res = fscanf (fp, " %s %d %d\n", format, vars, clauses);

    if (res == 0 || res == EOF)
    {
        fprintf (stderr, "Could not parse DIMACS SAT header\n");
        return 0;
    }

    if (strcmp ("cnf", format) != 0)
    {
        fprintf (stderr, "DIMACS header is not in CNF format\n");
        return 0;
    }

    return 1;
}

static int parse_Literal (char *buf)
/**************************************************************
 INPUT:     a parsed token
 RETURNS:   nonzero if the input has the form (+|-|)[1-9]*
***************************************************************/
{
    int c;

    c = 0;

    if (buf[c] == '-')
        c++;

    while (buf[c] != '\0')
    {
        if (!('0' <= buf[c] && buf[c] <= '9'))
            return 0;

        c++;
    }

    return 1;
}

int parse_Parse (FILE *fp, LIST **cls, int *num_vars, int *num_clauses)
/**************************************************************
 INPUT:     file handle of file to be parsed
            pointer to a pointer that will store an array of
            clauses on successful
            pointers to integers that will store number of
            variables/clauses on success
 RETURNS:   nonzero if parsing was successful
 MEMORY:    *cls will store the address of an array (if the
            return value is nonzero), this array must be freed
            manually
 SUMMARY:   parse a DIMACS SAT cnf file
***************************************************************/
{
    int input;              /* used for reading from file                               */
    char buffer[512];       /* buffer for reading token from file                       */
    int pos;                /* position in buffer                                       */
    int saw_header;         /* nonzero if header has been parsed                        */
    int vars, clauses;      /* information from file header                             */
    LIST *clauses_array;    /* a (dynamically allocated) array of lists of literals     */
    int curr_clause;        /* clause currently read                                    */

    pos = 0;
    saw_header = 0;
    curr_clause = 0;
    clauses_array = NULL;

    /* set the values we got pointers for to sane values */
    *cls = NULL;
    *num_vars = 0;
    *num_clauses = 0;

    do
    {
        /* guard for buffer overflow with very large token */
        if (pos >= 512)
        {
            fprintf (stderr, "Encountered a token that was too long\n");
            return 0;
        }

        input = fgetc (fp);
        buffer[pos] = (char) input;

        if (buffer[pos] == ' ' || buffer[pos] == '\t' || buffer[pos] == '\n' ||
            buffer[pos] == '\r' || buffer[pos] == EOF)
        {
            /* shortcut for skipping whitespace */
            if (pos == 0)
                continue;

            /* white spaces are token delimiters */
            buffer[pos] = '\0';

            /* reset position in buffer for next token */
            pos = 0;
        }
        else
        {
            /* keep parsing token */
            pos++;
            continue;
        }

        /* at this point the buffer stores a complete token */
        
        if (buffer[0] == 'c')
        {
            /* skip this line (if it contains something) */
            if (input != '\n')
                parse_Skip (fp);
        }
        else if (buffer[0] == 'p')
        {
            /* check if we already saw a header */
            if (saw_header)
            {
                fprintf (stderr, "Saw second header\n");
                free (clauses_array);
                return 0;
            }

            /* parse header and check file format */
            if (!parse_Header (fp, &vars, &clauses))
                return 0;

            clauses_array = (LIST*) malloc (sizeof (LIST) * clauses);

            if (!clauses_array)
            {
                fprintf (stderr, "Memory allocation error\n");
                return 0;
            }

            memset (clauses_array, 0, sizeof (LIST) * clauses);
            saw_header = 1;
        }
        else if (buffer[0] == '0')
        {
            /* clause was read completely */
            curr_clause++;

            /* check if we still/already can read a zero */
            if (!saw_header || curr_clause > clauses)
            {
                fprintf (stderr, "Unexpected zero token\n");

                if (saw_header)
                    free (clauses_array);

                return 0;
            }
        }
        else if (buffer[0] != '\0')
        {
            long literal;
            long id;

            /* must see header before any literals */
            if (!saw_header)
            {
                fprintf (stderr, "Error parsing DIMACS SAT header\n");
                return 0;
            }

            /* only as many clauses as specified in header */
            if (curr_clause >= clauses)
            {
                fprintf (stderr, "Too many clauses, reject token %s and all following\n", buffer);
                free (clauses_array);
                return 0;
            }

            /* must be a valid literal */
            if (!parse_Literal (buffer))
            {
                fprintf (stderr, "Invalid token %s\n", buffer);
                free (clauses_array);
                return 0;
            }

            /* get literal from buffer */
            sscanf (buffer, "%ld", &literal);

            /* get id of variable in literal */
            if (literal < 0)
                id = -literal;
            else id = literal;

            /* check if id is bigger than announced */
            if (id > vars)
            {
                fprintf (stderr, "Invalid variable %ld (max. allowed is %d)\n", id, vars);
                free (clauses_array);
                return 0;
            }

            /* add literal to clause */
            clauses_array[curr_clause] = list_Prepend ((void*) literal, clauses_array[curr_clause]);
        }
    }
    while (!feof (fp) || pos != 0);     /* parse as long as there is some unprocessed data */


    /* store the number of vars/clauses in the argument pointers */
    *num_vars = vars;
    *num_clauses = curr_clause;     /* use number of clauses we really, not number from header */

    /* return pointer to array */
    *cls = clauses_array;

    return 1;
}
